/*
 * Perform directory operation(s)
 *
 * -f name	Basename of file, slashes allowed
 * -n count	Create count files (name.0 name.1 ....). Default 100
 * -r nr	Repeat nr times; -1 means infinitely
 * -u uid	Use uid to create files
 * -U uid	Expect created files to have uid
 * -g gid	Use gid to create files
 * -G gid	Expect created files to have gid
 * -m mode	Use mode to create files
 * -M mode	Expect created files to have mode
 * -s		Sync file system after each pass
 * -x mask	Set umask to mask. Default 0.
 * -z		Do not fail creat/mkdir if file exists, try operation
 *		nevertheless.
 * -t delay	Sleep for delay seconds after each pass.
 * -T delay	Sleep for delay seconds before starting.
 * operation	One or more of creat unlink mkdir or rmdir.
 *		If specifying e.g. `creat mkdir' the program will
 *		first create the number of dirs specified by -n,
 *		then remove them.
 *
 * Copyright (C) 1998, Olaf Kirch <okir@monad.swb.de>
 */
#include <sys/types.h>
#include <sys/stat.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

static int		createuid,
			checkuid = -1,
			creategid,
			checkgid = -1,
			createmode = 0644,
			checkmode = -1,
			failifexists = 1;
#define OP_CREATE	1
#define OP_MKDIR	2
#define OP_UNLINK	3
#define OP_RMDIR	4
#define IS_REMOVEOP(o)	(((o) == OP_UNLINK) || ((o) == OP_RMDIR))


static void	perform(const char *filename, int op, int createmode);

static void
usage(int code)
{
	printf("usage:\n"
		"dirop [-f filename] [-n nrfiles] [-r repeat] "
				"[-t delay] [-T initdelay]\n"
		" [-g gid] [-G checkgid] [-u uid] [-U checkuid] "
				"[-m mode] [-M checkmode]\n"
		" [-x umask] [-s] [-z] operation ...\n");
	exit(code);
}

static void
pfail(const char *msg, const char *filename)
{
	printf("%s: %s failed: %s\n", filename, msg, strerror(errno));
	exit(1);
}

static void
fail(const char *msg, const char *filename)
{
	printf("%s: %s\n", filename, msg);
	exit(1);
}

static int
getnum(const char *string)
{
	char	*end;
	int	val;

	val = strtol(string, &end, 0);
	if (*end) {
		printf("Invalid number %s\n", string);
		exit(1);
	}
	return val;
}

int
main(int argc, char **argv)
{
	char	*basename = "file", *filename;
	int	initdelay = 0,
		delay = 0,
		nrfiles = 100,
		sumask = 0,
		repeat = 0,
		osync = 0,
		op[16], nrops, i, j, c;

	createuid = geteuid();
	creategid = getegid();

	while ((c = getopt(argc, argv, "f:G:g:M:m:n:r:T:t:sU:u:hx:z")) != -1) {
		switch (c) {
		case 'f':
			basename = optarg;
			break;
		case 'g':
			creategid = getnum(optarg);
			break;
		case 'G':
			checkgid = getnum(optarg);
			break;
		case 'm':
			createmode = getnum(optarg);
			break;
		case 'M':
			checkmode = getnum(optarg);
			break;
		case 'n':
			nrfiles = getnum(optarg);
			break;
		case 'r':
			repeat = getnum(optarg);
			break;
		case 's':
			osync++;
			break;
		case 't':
			delay = getnum(optarg);
			break;
		case 'T':
			initdelay = getnum(optarg);
			break;
		case 'u':
			createuid = getnum(optarg);
			break;
		case 'U':
			checkuid = getnum(optarg);
			break;
		case 'x':
			sumask = getnum(optarg);
			break;
		case 'z':
			failifexists = 0;
			break;
		case 'h':
			usage(0);
		case '?':
			usage(1);
		}
	}

	if (optind == argc) {
		printf("Expected operation mode.\n");
		usage(1);
	}

	nrops = 0;
	while (optind < argc) {
		optarg = argv[optind++];
		if (!strcmp(optarg, "creat"))
			op[nrops++] = OP_CREATE;
		else if (!strcmp(optarg, "mkdir"))
			op[nrops++] = OP_MKDIR;
		else if (!strcmp(optarg, "unlink"))
			op[nrops++] = OP_UNLINK;
		else if (!strcmp(optarg, "rmdir"))
			op[nrops++] = OP_RMDIR;
		else {
			printf("Unknown operation mode %s\n", optarg);
			usage(1);
		}
	}

	if (creategid != getegid()) {
		if (setegid(creategid) < 0) {
			perror("setegid");
			printf("Sorry, can't run this test under the "
				"requested gid -- skipping.\n");
			exit(0);
		}
	}

	if (createuid != geteuid()) {
		if (seteuid(createuid) < 0) {
			perror("seteuid");
			printf("Sorry, can't run this test under the "
				"requested uid -- skipping.\n");
			exit(0);
		}
	}

	umask(sumask);

	if (checkuid < 0)
		checkuid = createuid;
	if (checkgid < 0)
		checkgid = creategid;
	if (checkmode < 0)
		checkmode = createmode & ~sumask;
	if (repeat == 0)
		repeat = 1;
	filename = (char *) malloc(strlen(basename) + 32);

	if (initdelay)
		sleep(initdelay);

	while (1) {
		for (i = 0; i < nrops; i++) {
			for (j = 0; j < nrfiles; j++) {
				sprintf(filename, "%s.%d", basename, j);
				perform(filename, op[i], createmode);
			}
		}
		if (repeat > 0 && --repeat == 0)
			break;
		if (osync)
			sync();
		if (delay)
			sleep(delay);
	}

	return 0;
}

static void
perform(const char *filename, int op, int createmode)
{
	struct stat	pre_stat, post_stat;
	int		fd;

	if (stat(filename, &pre_stat) < 0) {
		pre_stat.st_ino = 0;
		if (errno != ENOENT || IS_REMOVEOP(op))
			pfail("pre-op stat", filename);
	} else {
		if (pre_stat.st_ino == 0)
			fail("zero inode #", filename);
		if (pre_stat.st_nlink == 0)
			fail("zero link count", filename);
		if (!IS_REMOVEOP(op) && failifexists)
			fail("exists\n", filename);
	}

	switch (op) {
	case OP_CREATE:
		fd = creat(filename, createmode);
		if (fd < 0)
			pfail("creat", filename);
		close(fd);
		break;
	case OP_MKDIR:
		if (mkdir(filename, createmode) < 0)
			pfail("mkdir", filename);
		break;
	case OP_UNLINK:
		if (unlink(filename) < 0)
			pfail("unlink", filename);
		break;
	case OP_RMDIR:
		if (rmdir(filename) < 0)
			pfail("rmdir", filename);
		break;
	}

	if (stat(filename, &post_stat) < 0) {
		post_stat.st_ino = 0;
		if (errno != ENOENT || !IS_REMOVEOP(op))
			pfail("post-op stat", filename);
	} else {
		if (post_stat.st_ino == 0)
			fail("zero inode #", filename);
		if (post_stat.st_nlink == 0)
			fail("zero link count", filename);
		if (IS_REMOVEOP(op))
			fail("still exists\n", filename);
		else
		if (pre_stat.st_ino) {
			/* This can be used to verify the operation
			 * of CREAT on an existing file */
			if (pre_stat.st_mode != post_stat.st_mode) {
				printf("%s: mode change 0%o -> 0%o\n",
					filename,
					pre_stat.st_mode,
					post_stat.st_mode);
				exit(1);
			}
			if (pre_stat.st_nlink != post_stat.st_nlink) {
				printf("%s: link # change %d -> %d\n",
					filename,
					pre_stat.st_nlink,
					post_stat.st_nlink);
				exit(1);
			}
			if (pre_stat.st_uid != post_stat.st_uid) {
				printf("%s: uid change %d -> %d\n",
					filename,
					pre_stat.st_uid,
					post_stat.st_uid);
				exit(1);
			}
			if (pre_stat.st_gid != post_stat.st_gid) {
				printf("%s: gid change %d -> %d\n",
					filename,
					pre_stat.st_gid,
					post_stat.st_gid);
				exit(1);
			}
		} else {
			if ((post_stat.st_mode & ~S_IFMT) != checkmode) {
				printf("%s: bad mode 0%o, wanted 0%o\n",
					filename,
					post_stat.st_mode & ~S_IFMT,
					createmode);
				exit(1);
			}
			if (post_stat.st_uid != checkuid) {
				printf("%s: bad uid %d, expected %d\n",
					filename, post_stat.st_uid, checkuid);
				exit(1);
			}
			if (post_stat.st_gid != checkgid) {
				printf("%s: bad gid %d, expected %d\n",
					filename, post_stat.st_gid, checkgid);
				exit(1);
			}
		}
	}
}
