/*
 * Write to a file.
 *
 * -o offset	Start writing at offset.
 * -c count	Write count bytes
 * -i incr	Increment incr bytes after writing count bytes
 *		(relative to file position after writing)
 * -l maxsize	Start over at offset when reaching maxsize (default 16m)
 * -r repeat	Do that repeat times (default -1 = forever)
 * -t offset	Truncate file to offset after reaching maxsize
 * -d delay	Sleep for delay seconds after each write
 * -s		sync file after each write
 * -ss		sync the file system after each write
 * -sss		open the file with O_SYNC
 * -R		reopen file after reaching maxsize
 * -v		Verify file after reaching maxsize
 * -z		truncate file before starting
 * -f name	Specify test file, - for stdin.
 *
 * All numeric values (except the delay) can be specified with an
 * optional unit of k, m, or g, meaning kilo-, mega- or gigabyte.
 */
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <string.h>
#include <ctype.h>
#include <fcntl.h>

/* Must be power of 2 */
#define CHUNK		64 * 1024

static long		getnum(const char *);

static void
usage(int code)
{
	printf("usage:\n"
		"[-o beginoffset] [-c count] [-i increment] [-l maxsize] "
		"[-r repeat] [-t truncoffset] "
		"[-d delay] [-s] [-z] [-f filename]\n");
	exit(code);
}

int
main(int argc, char **argv)
{
	const char	*fname = "junk";
	char		buffer[CHUNK], check[CHUNK];
	int		count = 4096,
			length = 4096 * 4096,
			increment = 0,
			offset = 0,
			delay = 0,
			repeat = -1,
			otrunc = 0,
			truncate = -1,
			osync = 0,
			overify = 0,
			oreopen = 0,
			openmode = O_RDWR|O_CREAT,
			c, fd = -1, where, left, i, j, verifying;

	while ((c = getopt(argc, argv, "c:d:f:i:l:o:Rr:st:vz")) != -1) {
		switch (c) {
		case 'c':
			count = getnum(optarg);
			break;
		case 'd':
			delay = atoi(optarg);
			break;
		case 'f':
			fname = optarg;
			break;
		case 'i':
			increment = getnum(optarg);
			break;
		case 'l':
			length = getnum(optarg);
			break;
		case 'o':
			offset = getnum(optarg);
			break;
		case 'R':
			oreopen = 1;
			break;
		case 'r':
			repeat = getnum(optarg);
			break;
		case 's':
			osync++;
			break;
		case 't':
			truncate = getnum(optarg);
			break;
		case 'v':
			overify = 1;
			break;
		case 'z':
			otrunc = 1;
			break;
		case 'h':
			usage(0);
		default:
			usage(1);
		}
	}

	if (!strcmp(fname, "-")) {
		fd = 0;
	} else {
		if (otrunc)
			openmode |= O_TRUNC;
		if (osync >= 3)
			openmode |= O_SYNC;
	}

	/* init buffer */
	memcpy(buffer, "0123456789abcdefghijklmnopqrstuv", 32);
	for (i = 32; i < CHUNK; i <<= 1)
		memcpy(buffer + i, buffer, i);

	if (repeat == 0)
		repeat = 1;

	verifying = 0;
	where = offset;
	while (1) {
	again:
		if (fd < 0 && (fd = open(fname, openmode, 0666)) < 0) {
			perror(fname);
			exit(1);
		}
		if (lseek(fd, where, SEEK_SET) < 0) {
			perror("lseek");
			exit(1);
		}
		left = count;
		i = 0;
		while (left) {
			j = left;
			if (i + j > CHUNK)
				j = i - CHUNK;
			if (!verifying)
				j = write(fd, buffer + i, j);
			else
				j = read(fd, check + i, j);
			if (j < 0) {
				perror("write");
				exit(1);
			}
			if (verifying) {
				if (memcmp(buffer + i, check + i, j)) {
					printf("File mismatch in %d-%d\n",
						offset, offset + count);
					exit(1);
				}
			}
			if ((i += j) >= CHUNK)
				i = 0;
			left -= j;
		}

		if (osync == 1) {
			if (fsync(fd) < 0) {
				perror("fsync");
				exit(1);
			}
		} else if (osync == 2) {
			if (sync() < 0) {
				perror("sync");
				exit(1);
			}
		}

		where = lseek(fd, 0, SEEK_CUR);
		if (increment)
			where += increment;
		if (delay)
			sleep(delay);
		if (where >= length) {
			where = offset;
			if (overify) {
				verifying = !verifying;;
				if (verifying)
					goto again;
			}
			if (repeat > 0 && --repeat == 0)
				break;
			if (truncate >= 0) {
				if (ftruncate(fd, truncate) < 0) {
					perror("truncate");
					exit(1);
				}
			}
			if (oreopen) {
				if (close(fd) < 0) {
					perror("close");
					exit(1);
				}
				fd = -1;
			}
		}
	}
	return 0;
}

static long
getnum(const char *string)
{
	char	*end;
	long	val;

	val = strtol(string, &end, 0);
	if (*end == 'k' || *end == 'K')
		val *= 1024;
	else if (*end == 'm' || *end == 'M')
		val *= 1024 * 1024;
	else if (*end == 'g' || *end == 'G')
		val *= 1024 * 1024 * 1024;
	else if (*end == '\0')
		return val;
	if (*++end != '\0') {
		printf("Invalid number \"%s\"\n", string);
		exit(1);
	}

	return val;
}
